<?php

if( !class_exists( 'VIP_GO_Thumbnail_Editor' ) ) {
    require_once __DIR__ . '/vip-go-thumbnail-editor.php';
}

/**
 * Adding new method for multiple thumbnail cropping, using VIP_GO_Thumbnail_Editor
 */
class DMG_Thumbnail_Editor extends VIP_GO_Thumbnail_Editor {

	/**
	 * DMG_Thumbnail_Editor constructor.
	 */
	function __construct() {
		if ( is_admin() ) {
			// Add a new field to the edit attachment screen, set at 50 to move above all thumbnails
			add_filter( 'attachment_fields_to_edit', [ $this, 'add_attachment_fields_to_edit_all' ], 50, 2 );

			// Create a new screen for editing a thumbnail
			add_action( 'admin_action_vip_go_thumbnail_edit_all', [ $this, 'show_all_thumbnails' ] );

			// Handle the form submit of the edit multiple thumbnail screen
			add_action( 'admin_post_vip_go_thumbnail_edit_all', [ $this, 'post_handler_all' ] );
		}
	}

	/**
	 * Adds a new field to add the "Edit All Thumbnails" button
	 *
	 * @param $form_fields
	 * @param $attachment
	 *
	 * @return mixed
	 */
	public function add_attachment_fields_to_edit_all( $form_fields, $attachment ) {
		if ( ! wp_attachment_is_image( $attachment->ID ) ) {
			return $form_fields;
		}

		$form_fields['vip_go_thumbnails_all'] = [
			'input' => 'html',
			'html'  => $this->get_attachment_field_all_html( $attachment ),
		];

		return $form_fields;
	}

	/**
	 * Generates the HTML for the "Edit All Thumbnails" button
	 *
	 * @param $attachment
	 *
	 * @return string
	 */
	public function get_attachment_field_all_html( $attachment ) {
		$html = '<input type="button" class="hide-if-no-js button" onclick="window.open(\'' . esc_url( admin_url( 'admin.php?action=vip_go_thumbnail_edit_all&id=' . intval( $attachment->ID ) . '&size=all' ) ) . '\');" value="' . __( 'Edit All Thumbnails', 'vip-go-thumbnail-editor' ) . '" />';

		return $html;
	}

	/**
	 * Outputs the HTML for the multiple thumbnail crop edit screen.
	 *
	 * @param $attachment
	 */
	public function show_all_thumbnails( $attachment ) {

		global $parent_file, $submenu_file, $title;

		// Validate "id" and "size" query string values and check user capabilities. Dies on error.
		$attachment = $this->validate_parameters_all();

		if ( empty( $_REQUEST['size'] ) || 'all' !== sanitize_text_field( $_REQUEST['size'] ) ) { // input var okay. validation okay. sanitization okay. Validated in this->validate_parameters_all()
			wp_die( 'Invalid size parameter' );
		}

		$sizes = $GLOBALS['Vip_Go_Thumbnail_Editor']->use_ratio_map ? $GLOBALS['Vip_Go_Thumbnail_Editor']->get_image_sizes_by_ratio() : $GLOBALS['Vip_Go_Thumbnail_Editor']->get_intermediate_image_sizes();

		$parent_file  = 'upload.php';
		$submenu_file = 'upload.php';

		$title = sprintf( esc_html__( 'Edit All Thumbnails', 'vip-go-thumbnail-editor' ) );

		wp_enqueue_script( 'imgareaselect' );
		wp_enqueue_style( 'imgareaselect' );

        // @ TODO remove jQuery
		wp_enqueue_script(
			'dmg-thumbnail-editor',
			plugin_dir_url( __FILE__ ) . '/dmg-thumbnail-editor.js',
			[
				'jquery',
			],
			null,
			true
		);

		require( ABSPATH . '/wp-admin/admin-header.php' );
		?>
		<div class="wrap">
			<form action="<?php echo esc_url( admin_url( 'admin-post.php' ) ); ?>" method="post">
				<?php screen_icon(); ?>
				<h2><?php echo esc_html( $title ); ?></h2>
				<?php
				$all_sizes = [];
				foreach ( $sizes as $key => $size ) {
					$image = image_downsize( $attachment->ID, [ 512, 512 ] );
					// Make sure the image fits on the screen
					if ( ! $image ) {
						wp_die( esc_html__( 'Failed to downsize the original image to fit on your screen. How odd. Please contact support.', 'vip-go-thumbnail-editor' ) );
					}

					$thumbnail_dimensions = $GLOBALS['Vip_Go_Thumbnail_Editor']->get_thumbnail_dimensions( $size );
					// How big is the final thumbnail image?
					if ( ! $thumbnail_dimensions ) {
						wp_die( 'Invalid <code>size</code> parameter.' );
					}

					$original_aspect_ratio  = $image[1] / $image[2];
					$thumbnail_aspect_ratio = $thumbnail_dimensions['width'] / $thumbnail_dimensions['height'];
					$initial_selection      = [];

					# Build the selection coordinates

					$coordinates = $GLOBALS['Vip_Go_Thumbnail_Editor']->get_coordinates( $attachment->ID, $size );
					// If there's already a custom selection
					if ( $coordinates ) {
						$attachment_metadata = wp_get_attachment_metadata( $attachment->ID );
						// If original is bigger than display, scale down the coordinates to match the scaled down original
						if ( $attachment_metadata['width'] > $image[1] || $attachment_metadata['height'] > $image[2] ) {

							// At what percentage is the image being displayed at?
							$scale = $image[1] / $attachment_metadata['width'];

							foreach ( $coordinates as $coordinate ) {
								$initial_selection[] = round( $coordinate * $scale );
							}
						} // Or the image was not downscaled, so the coordinates are correct
						else {
							$initial_selection = $coordinates;
						}
					} // If original and thumb are the same aspect ratio, then select the whole image
					elseif ( $thumbnail_aspect_ratio === $original_aspect_ratio ) {
						$initial_selection = [ 0, 0, $image[1], $image[2] ];
					} // If the thumbnail is wider than the original, we want the full width
					elseif ( $thumbnail_aspect_ratio > $original_aspect_ratio ) {
						// Take the width and divide by the thumbnail's aspect ratio
						$selected_height = round( $image[1] / ( $thumbnail_dimensions['width'] / $thumbnail_dimensions['height'] ) );

						$initial_selection = [
							0,
							// Far left edge (due to aspect ratio comparison)
							round( ( $image[2] / 2 ) - ( $selected_height / 2 ) ),
							// Mid-point + half of height of selection
							$image[1],
							// Far right edge (due to aspect ratio comparison)
							round( ( $image[2] / 2 ) + ( $selected_height / 2 ) ),
							// Mid-point - half of height of selection
						];
					} // The thumbnail must be narrower than the original, so we want the full height
					else {
						// Take the width and divide by the thumbnail's aspect ratio
						$selected_width = round( $image[2] / ( $thumbnail_dimensions['height'] / $thumbnail_dimensions['width'] ) );

						$initial_selection = [
							round( ( $image[1] / 2 ) - ( $selected_width / 2 ) ),
							// Mid-point + half of height of selection
							0,
							// Top edge (due to aspect ratio comparison)
							round( ( $image[1] / 2 ) + ( $selected_width / 2 ) ),
							// Mid-point - half of height of selection
							$image[2],
							// Bottom edge (due to aspect ratio comparison)
						];
					}

					?>
					<div class="postbox">
						<div class="inside">
							<h3 class="hndle">
							<?php
							/* translators: %s: image size  */
								printf( esc_html__( '%s Thumbnail Preview', 'vip-go-thumbnail-editor' ), esc_html( $key ) );
							?>
								</h3>

							<p>
								<img src="<?php echo esc_url( $image[0] ); ?>" width="<?php echo (int) $image[1]; ?>" height="<?php echo (int) $image[2]; ?>" id="<?php /* translators: %s: image size  */ echo esc_attr( 'vip-go-thumbnail-edit-' . $size ); ?>" alt="<?php echo esc_attr( sprintf( esc_attr__( '%s Thumbnail', 'vip-go-thumbnail-editor' ), $size ) ); ?>" />
							</p>
							<div style="overflow:hidden;width:<?php echo (int) $thumbnail_dimensions['width']; ?>px;height:<?php echo (int) $thumbnail_dimensions['height']; ?>px;">
								<img id="<?php echo esc_attr( 'vip-go-thumbnail-edit-preview-' . $size ); ?>" class="hidden" src="<?php echo esc_url( wp_get_attachment_url( $attachment->ID ) ); ?>" />
							</div>
						</div>
					</div>
					<!--
					Since the fullsize image is possibly scaled down, we need to record at what size it was
					displayed at so the we can scale up the new selection dimensions to the fullsize image.
					-->
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_display_width" value="<?php echo (int) $image[1]; ?>" />
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_display_height" value="<?php echo (int) $image[2]; ?>" />
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_width" value="<?php echo (int) $thumbnail_dimensions['width']; ?>" />
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_height" value="<?php echo (int) $thumbnail_dimensions['height']; ?>" />

					<!-- These are manipulated via Javascript to submit the selected values -->
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_x1" value="<?php echo (int) $initial_selection[0]; ?>" />
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_y1" value="<?php echo (int) $initial_selection[1]; ?>" />
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_x2" value="<?php echo (int) $initial_selection[2]; ?>" />
					<input type="hidden" name="<?php echo esc_attr( 'vip_go_thumbnail_edit_' . $size ); ?>_y2" value="<?php echo (int) $initial_selection[3]; ?>" />

					<?php
					$all_sizes_array[] = $size;
				}
				$all_sizes = implode( ',', $all_sizes_array );

				wp_localize_script(
					'dmg-thumbnail-editor',
					'params',
					[
						'all_sizes' => esc_js( $all_sizes ),
					]
				);

		?>
				<input type="hidden" name="action" value="vip_go_thumbnail_edit_all" />
				<input type="hidden" name="id" value="<?php echo esc_attr( (int) $attachment->ID ); ?>" />
				<input type="hidden" name="all_sizes" value="<?php echo esc_attr( $all_sizes ); ?>" />
				<input type="hidden" name="size" value="all" />
				<?php wp_nonce_field( 'vip_go_thumbnail_edit_' . $attachment->ID . '_all' ); ?>
				<p>
					<?php submit_button( null, 'primary', 'submit', false ); ?>
					<a href="<?php echo esc_url( admin_url( 'media.php?action=edit&attachment_id=' . $attachment->ID ) ); ?>" class="button"><?php esc_html_e( 'Cancel Changes', 'vip-go-thumbnail-editor' ); ?></a>
				</p>
			</form>

		</div>
		<?php
		require( ABSPATH . '/wp-admin/admin-footer.php' );

	}

	/**
	 * Processes the submission of the thumbnail crop selection screen and saves the results to post meta, for multiple images.
	 *
	 * @param $attachment
	 */
	public function post_handler_all( $attachment ) {

		// Validate "id" and "size" POST values and check user capabilities. Dies on error.
		$attachment = $this->validate_parameters_all();

		if ( ! isset( $_POST['_wpnonce'] ) || ! wp_verify_nonce( $_POST['_wpnonce'], 'vip_go_thumbnail_edit_' . $attachment->ID . '_all' ) ) { // input var okay. validation okay. sanitization okay.
			wp_die( 'Not authorized' );
		}

		check_admin_referer( 'vip_go_thumbnail_edit_' . $attachment->ID . '_all' );

		if ( ! empty( $_POST['all_sizes'] ) && is_array( $all_sizes = explode( ',', sanitize_text_field( wp_unslash( $_POST['all_sizes'] ) ) ) ) ) { // Input var okay.

			foreach ( $all_sizes as $size ) {

				$required_fields = [
					'vip_go_thumbnail_edit_' . $size . '_display_width' => 'display_width',
					'vip_go_thumbnail_edit_' . $size . '_display_height' => 'display_height',
					'vip_go_thumbnail_edit_' . $size . '_x1' => 'selection_x1',
					'vip_go_thumbnail_edit_' . $size . '_y1' => 'selection_y1',
					'vip_go_thumbnail_edit_' . $size . '_x2' => 'selection_x2',
					'vip_go_thumbnail_edit_' . $size . '_y2' => 'selection_y2',
				];

				foreach ( $required_fields as $required_field => $variable_name ) {
					if ( empty( $_POST[ $required_field ] ) && 0 !== (int) $_POST[ $required_field ] ) { // Input var okay.
						/* translators: %s: required field  */
						wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>' . esc_html( $required_field ) . '</code>' ) );
					}

					$$variable_name = (int) $_POST[ $required_field ]; // Input var okay.
				}

				$attachment_metadata = wp_get_attachment_metadata( $attachment->ID );

				$selection_coordinates = [ 'selection_x1', 'selection_y1', 'selection_x2', 'selection_y2' ];

				// If the image was scaled down on the selection screen,
				// then we need to scale up the selection to fit the fullsize image
				if ( $attachment_metadata['width'] > $display_width || $attachment_metadata['height'] > $display_height ) {
					$scale_ratio = $attachment_metadata['width'] / $display_width;

					foreach ( $selection_coordinates as $selection_coordinate ) {
						${'fullsize_' . $selection_coordinate} = round( $$selection_coordinate * $scale_ratio );
					}
				} else {
					// Remap
					foreach ( $selection_coordinates as $selection_coordinate ) {
						${'fullsize_' . $selection_coordinate} = $$selection_coordinate;
					}
				}
				// Save the coordinates
				$GLOBALS['Vip_Go_Thumbnail_Editor']->save_coordinates(
					$attachment->ID,
					$size,
					[
						$fullsize_selection_x1,
						$fullsize_selection_y1,
						$fullsize_selection_x2,
						$fullsize_selection_y2,
					]
				);
			}
		}

		wp_safe_redirect( admin_url( 'media.php?action=edit&attachment_id=' . $attachment->ID . '&wteupdated=1' ) );
		exit();
	}

	/**
	 * Makes sure that the "id" (attachment ID) and "size" (thumbnail size) query string parameters are valid
	 * and dies if they are not. Returns attachment object with matching ID on success.
	 *
	 * @return null|object Dies on error, returns attachment object on success.
	 */
	public function validate_parameters_all() {

		if ( empty( $_REQUEST['id'] ) || ! $attachment = get_post( intval( $_REQUEST['id'] ) ) ) { // Input var okay.
			/* translators: %s: html string */
			wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>id</code>' ) );
		}

		if ( 'attachment' !== $attachment->post_type || ! wp_attachment_is_image( $attachment->ID ) ) {
			wp_die( esc_html__( 'That is not a valid image attachment.', 'vip-go-thumbnail-editor' ) );
		}

		if ( ! current_user_can( get_post_type_object( $attachment->post_type )->cap->edit_post, $attachment->ID ) ) {
			wp_die( esc_html( 'You are not allowed to edit this attachment.' ) );
		}

		if ( empty( $_REQUEST['size'] ) || 'all' !== $_REQUEST['size'] ) { // Input var okay.
			/* translators: %s: html string */
			wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>size</code>' ) );
		}

		return $attachment;
	}
}
