<?php

/**
 * Class VIP_GO_Thumbnail_Editor
 *
 * Lifted from the WordPress.com thumbnail editor plugin
 *
 * @see https://github.com/Automattic/wpcom-thumbnail-editor
 */
class VIP_GO_Thumbnail_Editor {

	/**
	 * Post meta key name, for storing crop coordinates.
	 * we are keeping the same post_meta name used by the wpcom version of the plugin to ensure the compatibility between them
	 *
	 * @var string
	 */
	public $post_meta = 'wpcom_thumbnail_edit';

	/**
	 * Determine if we want to use a dimension map or not
	 */
	public $use_ratio_map = false;

	/**
	 * Array used to validate the image sizes
	 */
	public $additional_image_sizes = [];

	/**
	 * An array that maps specific aspect ratios to image size strings. Should be defined by user to be accurate.
	 */
	public $image_ratio_map = [];

	/**
	 * Default settings for allowing private blogs to use this plugin.
	 */
	public $allow_private_blogs = false;

	/**
	 * Initialize the class by registering various hooks.
	 */
	function __construct() {
		global $_wp_additional_image_sizes;
		$args = apply_filters(
			'vip_go_thumbnail_editor_args', [
				'image_ratio_map' => false,
			]
		);

		// Allow for private blogs to use this plugin
		$this->allow_private_blogs = apply_filters( 'vip_go_thumbnail_editor_allow_private_blogs', $this->allow_private_blogs );

		// When a thumbnail is requested, intercept the request and return the custom thumbnail
		if ( ! function_exists( 'is_private_blog' ) || ( function_exists( 'is_private_blog' ) && ( ! is_private_blog() || true === $this->allow_private_blogs ) ) ) {
			add_filter( 'image_downsize', array( $this, 'get_thumbnail_url' ), 15, 3 );
		}

		// Admin-only hooks
		if ( is_admin() ) {

			// Add a new field to the edit attachment screen
			add_filter( 'attachment_fields_to_edit', [ $this, 'add_attachment_fields_to_edit' ], 50, 2 );

			// Create a new screen for editing a thumbnail
			add_action( 'admin_action_vip_go_thumbnail_edit', [ $this, 'edit_thumbnail_screen' ] );

			// Handle the form submit of the edit thumbnail screen
			add_action( 'admin_post_vip_go_thumbnail_edit', [ $this, 'post_handler' ] );

			// Display status messages
			add_action( 'admin_notices', [ $this, 'output_thumbnail_message' ] );

		}

		$this->additional_image_sizes = $_wp_additional_image_sizes;

		if ( is_array( $args['image_ratio_map'] ) ) {
			$this->use_ratio_map = true;

			// Validate image sizes

			foreach ( $args['image_ratio_map'] as $ratio => $image_sizes ) {
				$ratio_map[ $ratio ] = [];

				foreach ( $image_sizes as $image_size ) {
					if ( is_array( $this->additional_image_sizes ) && array_key_exists( $image_size, $this->additional_image_sizes ) ) {
						$ratio_map[ $ratio ][] = $image_size;
					}
				}

				if ( empty( $ratio_map[ $ratio ] ) ) {
					unset( $ratio_map[ $ratio ] );
				}
			}
			$this->image_ratio_map = $ratio_map;
		}

	}

	/**
	 * Outputs status messages based on query parameters.
	 *
	 * It cheats a little and uses the settings error API in order to avoid having to generate it's own HTML.
	 */
	public function output_thumbnail_message() {
		if ( isset( $_GET['wtereset'] ) && '1' === $_GET['wtereset'] ) { // Input var okay.
			add_settings_error( 'vip_go_thumbnail_edit', 'reset', esc_html__( 'Thumbnail position reset.', 'vip-go-thumbnail-editor' ), 'updated' );
		} elseif ( isset( $_GET['wteupdated'] ) && '1' === $_GET['wteupdated'] ) {  // Input var okay.
			add_settings_error( 'vip_go_thumbnail_edit', 'updated', esc_html__( 'Thumbnail position updated.', 'vip-go-thumbnail-editor' ), 'updated' );
		} else {
			return;
		}

		settings_errors( 'vip_go_thumbnail_edit' );
	}

	/**
	 * Adds a new field to the edit attachment screen that lists thumbnail sizes.
	 *
	 * @param $form_fields array Existing fields.
	 * @param $attachment object The attachment currently being edited.
	 *
	 * @return array Form fields, either unmodified on error or new field added on success.
	 */

	public function add_attachment_fields_to_edit( $form_fields, $attachment ) {

		if ( ! wp_attachment_is_image( $attachment->ID ) ) {
			return $form_fields;
		}

		$form_fields['vip_go_thumbnails'] = [
			'label' => 'Thumbnail Images',
			'input' => 'html',
			'html'  => $this->get_attachment_field_html( $attachment ),
		];

		return $form_fields;
	}

	/**
	 * Generates the HTML for the edit attachment field.
	 *
	 * @param $attachment object The attachment currently being edited.
	 *
	 * @return string The HTML for the form field.
	 */
	public function get_attachment_field_html( $attachment ) {

		$sizes = $this->use_ratio_map ? $this->get_image_sizes_by_ratio() : $this->get_intermediate_image_sizes();

		$sizes = apply_filters( 'vip_go_thumbnail_editor_image_size_names_choose', $sizes );

		if ( empty( $sizes ) ) {
			return '<p>' . esc_html__( 'No thumbnail sizes could be found that are cropped. For now this functionality only supports cropped thumbnails.', 'vip-go-thumbnail-editor' ) . '</p>';
		}

		$html = '<p class="hide-if-js">' . esc_html__( 'You need to enable Javascript to use this functionality.', 'vip-go-thumbnail-editor' ) . '</p>';

		$html .= '<input type="button" class="hide-if-no-js button" onclick="jQuery(this).hide();jQuery(\'#' . esc_js( 'vip-go-thumbs-' . $attachment->ID ) . '\').slideDown(\'slow\');" value="' . __( 'Show Thumbnails', 'vip-go-thumbnail-editor' ) . '" />';

		$html .= '<div id="' . esc_attr( 'vip-go-thumbs-' . $attachment->ID ) . '" class="hidden">';

		$html .= '<p>' . esc_html__( 'Click on a thumbnail image to modify it. Each thumbnail has likely been scaled down in order to fit nicely into a grid.<br /><strong>Only thumbnails that are cropped are shown.</strong> Other sizes are hidden because they will be scaled to fit.', 'vip-go-thumbnail-editor' ) . '</p>';

		$html .= '<div>';

		// key wont really matter if its not using a dimension map
		foreach ( $sizes as $key => $size ) {

			$image_name = $this->use_ratio_map ? $key : $size;

			$edit_url = admin_url( 'admin.php?action=vip_go_thumbnail_edit&id=' . intval( $attachment->ID ) . '&size=' . rawurlencode( $size ) );

			// add an extra query var if were using a ratio map
			if ( $this->use_ratio_map ) {
				$edit_url = add_query_arg( 'ratio', $key, $edit_url );
			}

			// We need to get the fullsize thumbnail so that the cropping is properly done
			$thumbnail = image_downsize( $attachment->ID, $size );

			$thumbnail_url = $thumbnail[0];

			$html .= '<div style="float:left;margin:0 20px 20px 0;min-width:250px;">';
			$html .= '<a href="' . esc_url( $edit_url ) . '"';

			$screen = get_current_screen();

			if ( $screen && 'media' !== $screen->id ) {
				$html .= ' target="_blank"';
			}

			$html .= '>';
			$html .= '<strong>' . esc_html( $image_name ) . '</strong><br />';
			$html .= '<img src="' . esc_url( $thumbnail_url ) . '" alt="' . esc_attr( $size ) . '" />';
			$html .= '</a>';
			$html .= '</div>';
		}

		$html .= '</div>';
		$html .= '</div>';

		return $html;
	}

	/**
	 * Outputs the HTML for the thumbnail crop selection screen.
	 */
	public function edit_thumbnail_screen() {
		global $parent_file, $submenu_file, $title;

		$screen = get_current_screen();

		if ( 'admin' === $screen->id && isset( $_REQUEST['action'] ) && 'vip_go_thumbnail_edit' === $_REQUEST['action'] ) {  // input var okay. validation okay.

			// Validate "id" and "size" query string values and check user capabilities. Dies on error.
			$attachment = $this->validate_parameters();

			$size  = sanitize_text_field( $_REQUEST['size'] ); // input var okay. validation okay. sanitization okay. Validated in this::validate_parameters()
			$image = image_downsize( $attachment->ID, [ 1024, 1024 ] );

			// Make sure the image fits on the screen
			if ( ! $image ) {
				wp_die( esc_html_e( 'Failed to downsize the original image to fit on your screen. How odd. Please contact support.', 'vip-go-thumbnail-editor' ) );
			}

			$thumbnail_dimensions = $this->get_thumbnail_dimensions( $size );
			// How big is the final thumbnail image?
			if ( ! $thumbnail_dimensions ) {
				wp_die( esc_html_e( 'Invalid <code>size</code> parameter.', 'vip-go-thumbnail-editor' ) );
			}

			$parent_file  = 'upload.php';
			$submenu_file = 'upload.php';

			// adjust the image name if were using a ratio map
			$image_name = isset( $_REQUEST['ratio'] ) ? sanitize_text_field( wp_unslash( $_REQUEST['ratio'] ) ) : $size; // input var okay.
			/* translators: %s: image name  */
			$title = sprintf( esc_html__( 'Edit Thumbnail: %s', 'vip-go-thumbnail-editor' ), esc_html( $image_name ) );

			$original_aspect_ratio  = $image[1] / $image[2];
			$thumbnail_aspect_ratio = $thumbnail_dimensions['width'] / $thumbnail_dimensions['height'];

			# Build the selection coordinates

			$coordinates = $this->get_coordinates( $attachment->ID, $size );
			// If there's already a custom selection
			if ( $coordinates ) {
				$attachment_metadata = wp_get_attachment_metadata( $attachment->ID );

				// If original is bigger than display, scale down the coordinates to match the scaled down original
				if ( $attachment_metadata['width'] > $image[1] || $attachment_metadata['height'] > $image[2] ) {

					// At what percentage is the image being displayed at?
					$scale = $image[1] / $attachment_metadata['width'];

					foreach ( $coordinates as $coordinate ) {
						$initial_selection[] = round( $coordinate * $scale );
					}
				} // Or the image was not downscaled, so the coordinates are correct
				else {
					$initial_selection = $coordinates;
				}
			} // If original and thumb are the same aspect ratio, then select the whole image
			elseif ( $thumbnail_aspect_ratio === $original_aspect_ratio ) {
				$initial_selection = [ 0, 0, $image[1], $image[2] ];
			} // If the thumbnail is wider than the original, we want the full width
			elseif ( $thumbnail_aspect_ratio > $original_aspect_ratio ) {
				// Take the width and divide by the thumbnail's aspect ratio
				$selected_height = round( $image[1] / ( $thumbnail_dimensions['width'] / $thumbnail_dimensions['height'] ) );

				$initial_selection = [
					0, // Far left edge (due to aspect ratio comparison)
					round( ( $image[2] / 2 ) - ( $selected_height / 2 ) ), // Mid-point + half of height of selection
					$image[1], // Far right edge (due to aspect ratio comparison)
					round( ( $image[2] / 2 ) + ( $selected_height / 2 ) ), // Mid-point - half of height of selection
				];
			} // The thumbnail must be narrower than the original, so we want the full height
			else {
				// Take the width and divide by the thumbnail's aspect ratio
				$selected_width = round( $image[2] / ( $thumbnail_dimensions['height'] / $thumbnail_dimensions['width'] ) );

				$initial_selection = [
					round( ( $image[1] / 2 ) - ( $selected_width / 2 ) ),
					// Mid-point + half of height of selection
					0,
					// Top edge (due to aspect ratio comparison)
					round( ( $image[1] / 2 ) + ( $selected_width / 2 ) ),
					// Mid-point - half of height of selection
					$image[2],
					// Bottom edge (due to aspect ratio comparison)
				];
			}

			wp_enqueue_script( 'imgareaselect' );
			wp_enqueue_style( 'imgareaselect' );

			wp_enqueue_script(
				'vip-go-thumbnail-editor', plugin_dir_url( __FILE__ ) . 'js/vip-go-thumbnail-editor.js', [
					'jquery',
				], null, true
			);
			wp_localize_script(
				'vip-go-thumbnail-editor', 'params', [
					'image'                => $image,
					'thumbnail_dimensions' => $thumbnail_dimensions,
					'initial_selection'    => $initial_selection,
				]
			);

			require_once( ABSPATH . '/wp-admin/admin-header.php' );

			?>

			<div class="wrap">
				<form action="<?php echo esc_url( admin_url( 'admin-post.php' ) ); ?>" method="post">

					<?php screen_icon(); ?>
					<h2><?php echo esc_html( $title ); ?></h2>

					<noscript><p>
							<strong style="color:red;font-size:20px;"><?php esc_html_e( 'Please enable Javascript to use this page.', 'vip-go-thumbnail-editor' ); ?></strong>
						</p></noscript>

					<p><?php esc_html_e( 'The original image is shown in full below, although it may have been shrunk to fit on your screen. Please select the portion that you would like to use as the thumbnail image.', 'vip-go-thumbnail-editor' ); ?></p>

					<p>
						<img src="<?php echo esc_url( $image[0] ); ?>" width="<?php echo (int) $image[1]; ?>" height="<?php echo (int) $image[2]; ?>" id="vip-go-thumbnail-edit" alt="<?php /* translators: %s: thumbnail size */ esc_attr( sprintf( __( '"%s" Thumbnail', 'vip-go-thumbnail-editor' ), $size ) ); ?>" />
					</p>

					<?php do_action( 'vip_go_thumbnail_editor_edit_thumbnail_screen', $attachment->ID, $size ); ?>

					<p>
						<?php submit_button( null, 'primary', 'submit', false ); ?>
						<?php submit_button( esc_html__( 'Reset Thumbnail', 'vip-go-thumbnail-editor' ), 'primary', 'vip_go_thumbnail_edit_reset', false ); ?>
						<a href="<?php echo esc_url( admin_url( 'media.php?action=edit&attachment_id=' . $attachment->ID ) ); ?>" class="button"><?php esc_html_e( 'Cancel Changes', 'vip-go-thumbnail-editor' ); ?></a>
					</p>

					<h3><?php esc_html_e( 'Fullsize Thumbnail Preview', 'vip-go-thumbnail-editor' ); ?></h3>

					<div style="overflow:hidden;width:<?php echo (int) $thumbnail_dimensions['width']; ?>px;height:<?php echo (int) $thumbnail_dimensions['height']; ?>px;">
						<img id="vip-go-thumbnail-edit-preview" class="hidden" src="<?php echo esc_url( wp_get_attachment_url( $attachment->ID ) ); ?>" />
					</div>

					<input type="hidden" name="action" value="vip_go_thumbnail_edit" />
					<input type="hidden" name="id" value="<?php echo (int) $attachment->ID; ?>" />
					<input type="hidden" name="size" value="<?php echo esc_attr( $size ); ?>" />
					<?php wp_nonce_field( 'vip_go_thumbnail_edit_' . $attachment->ID . '_' . $size ); ?>

					<!--
						Since the fullsize image is possibly scaled down, we need to record at what size it was
						displayed at so the we can scale up the new selection dimensions to the fullsize image.
					-->
					<input type="hidden" name="vip_go_thumbnail_edit_display_width" value="<?php echo (int) $image[1]; ?>" />
					<input type="hidden" name="vip_go_thumbnail_edit_display_height" value="<?php echo (int) $image[2]; ?>" />

					<!-- These are manipulated via Javascript to submit the selected values -->
					<input type="hidden" name="vip_go_thumbnail_edit_x1" value="<?php echo (int) $initial_selection[0]; ?>" />
					<input type="hidden" name="vip_go_thumbnail_edit_y1" value="<?php echo (int) $initial_selection[1]; ?>" />
					<input type="hidden" name="vip_go_thumbnail_edit_x2" value="<?php echo (int) $initial_selection[2]; ?>" />
					<input type="hidden" name="vip_go_thumbnail_edit_y2" value="<?php echo (int) $initial_selection[3]; ?>" />
				</form>
			</div>

			<?php

			require( ABSPATH . '/wp-admin/admin-footer.php' );

		}
	}

	/**
	 * Processes the submission of the thumbnail crop selection screen and saves the results to post meta.
	 */
	public function post_handler() {

		// Validate "id" and "size" POST values and check user capabilities. Dies on error.
		$attachment = $this->validate_parameters();

		$size = sanitize_text_field( $_REQUEST['size'] ); // input var okay. validation okay. sanitization okay. Validated in this::validate_parameters()

		if ( ! isset( $_POST['_wpnonce'] ) || ! wp_verify_nonce( $_POST['_wpnonce'], 'vip_go_thumbnail_edit_' . $attachment->ID . '_' . $size ) ) { // input var okay. validation okay. sanitization okay.
			wp_die( 'Not authorized' );
		}

		check_admin_referer( 'vip_go_thumbnail_edit_' . $attachment->ID . '_' . $size );

		// Reset to default?
		if ( ! empty( $_POST['vip_go_thumbnail_edit_reset'] ) ) { // Input var okay.
			$this->delete_coordinates( $attachment->ID, $size );

			wp_safe_redirect( admin_url( 'media.php?action=edit&attachment_id=' . $attachment->ID . '&wtereset=1' ) );
			exit();
		}

		$required_fields = [
			'vip_go_thumbnail_edit_display_width'  => 'display_width',
			'vip_go_thumbnail_edit_display_height' => 'display_height',
			'vip_go_thumbnail_edit_x1'             => 'selection_x1',
			'vip_go_thumbnail_edit_y1'             => 'selection_y1',
			'vip_go_thumbnail_edit_x2'             => 'selection_x2',
			'vip_go_thumbnail_edit_y2'             => 'selection_y2',
		];

		foreach ( $required_fields as $required_field => $variable_name ) {
			if ( empty( $_POST[ $required_field ] ) && 0 !== (int) $_POST[ $required_field ] ) { // Input var okay.
				/* translators: %s: param name  */
				wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>' . esc_html( $required_field ) . '</code>' ) );
			}

			${$variable_name} = (int) $_POST[ $required_field ]; // Input var okay.
		}

		$attachment_metadata = wp_get_attachment_metadata( $attachment->ID );

		$selection_coordinates = [ 'selection_x1', 'selection_y1', 'selection_x2', 'selection_y2' ];

		// If the image was scaled down on the selection screen,
		// then we need to scale up the selection to fit the fullsize image
		if ( $attachment_metadata['width'] > $display_width || $attachment_metadata['height'] > $display_height ) {
			$scale_ratio = $attachment_metadata['width'] / $display_width;

			foreach ( $selection_coordinates as $selection_coordinate ) {
				${'fullsize_' . $selection_coordinate} = round( ${$selection_coordinate} * $scale_ratio );
			}
		} else {
			// Remap
			foreach ( $selection_coordinates as $selection_coordinate ) {
				${'fullsize_' . $selection_coordinate} = ${$selection_coordinate};
			}
		}

		// Save the coordinates
		$this->save_coordinates(
			$attachment->ID, $size, [
				$fullsize_selection_x1,
				$fullsize_selection_y1,
				$fullsize_selection_x2,
				$fullsize_selection_y2,
			]
		);

		// Allow for saving custom fields
		do_action( 'vip_go_thumbnail_editor_post_handler', $attachment->ID, $size );

		wp_safe_redirect( admin_url( 'media.php?action=edit&attachment_id=' . $attachment->ID . '&wteupdated=1' ) );
		exit();
	}

	/**
	 * Makes sure that the "id" (attachment ID) and "size" (thumbnail size) query string parameters are valid
	 * and dies if they are not. Returns attachment object with matching ID on success.
	 *
	 * @return null|object Dies on error, returns attachment object on success.
	 */
	public function validate_parameters() {

		if ( empty( $_REQUEST['id'] ) || ! $attachment = get_post( intval( $_REQUEST['id'] ) ) ) { // Input var okay.
			/* translators: %s: html string  */
			wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>id</code>' ) );
		}

		if ( 'attachment' !== $attachment->post_type || ! wp_attachment_is_image( $attachment->ID ) ) {
			/* translators: %s: html string */
			wp_die( esc_html_e( 'That is not a valid image attachment.', 'vip-go-thumbnail-editor' ) );
		}

		if ( ! current_user_can( 'upload_files' ) ) {
			wp_die( esc_html_e( 'You are not allowed to edit this attachment.', 'vip-go-thumbnail-editor' ) );
		}

		if ( $this->use_ratio_map ) {
			if ( empty( $_REQUEST['size'] ) || ! in_array( wp_unslash( $_REQUEST['size'] ), $this->get_image_sizes_by_ratio(), true ) ) { // Input var okay.
				/* translators: %s: html string */
				wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>size</code>' ) );
			}
		} else {
			if ( empty( $_REQUEST['size'] ) || ! in_array( wp_unslash( $_REQUEST['size'] ), $this->get_intermediate_image_sizes(), true ) ) { // Input var okay.
				/* translators: %s: html string */
				wp_die( sprintf( esc_html__( 'Invalid %s parameter.', 'vip-go-thumbnail-editor' ), '<code>size</code>' ) );
			}
		}

		return $attachment;
	}

	/**
	 * Returns all thumbnail size names. get_intermediate_image_sizes() is filtered to return an
	 * empty array on WP VIP GO so this function removes that filter, calls the function,
	 * and then re-adds the filter back onto the function.
	 *
	 * @param bool $cropped_only
	 *
	 * @return array An array of image size strings.
	 */
	public function get_intermediate_image_sizes( $cropped_only = true ) {

		# /wp-content/mu-plugins/a8c-files.php
		$had_filter = remove_filter( 'intermediate_image_sizes', 'wpcom_intermediate_sizes' );

		$sizes = get_intermediate_image_sizes();

		if ( $had_filter ) {
			add_filter( 'intermediate_image_sizes', 'wpcom_intermediate_sizes' );
		}

		if ( apply_filters( 'vip_go_thumbnail_editor_cropped_only', $cropped_only ) ) {
			$filtered_sizes = [];

			foreach ( $sizes as $size ) {
				switch ( $size ) {
					case 'thumbnail':
						if ( get_option( 'thumbnail_crop' ) ) {
							$filtered_sizes[] = $size;
						}
						break;

					case 'medium':
					case 'large':
						break;

					default:
						if ( ! empty( $this->additional_image_sizes[ $size ] ) && $this->additional_image_sizes[ $size ]['crop'] ) {
							$filtered_sizes[] = $size;
						}
				}
			}

			$sizes = $filtered_sizes;
		}

		return apply_filters( 'wpcom_thumbnail_editor_get_intermediate_image_sizes', $sizes, $cropped_only );
	}

	/**
	 * Gets the first size defined for each dimension. All images are assumed to be cropped
	 *
	 * @todo Add validation that image sizes are of the cropped variety?
	 * @return array Array of image size strings.
	 */
	public function get_image_sizes_by_ratio() {

		$ratios = array_keys( $this->image_ratio_map );

		foreach ( $ratios as $ratio ) {
			if ( isset( $this->image_ratio_map[ $ratio ][0] ) ) {
				$sizes[ $ratio ] = $this->image_ratio_map[ $ratio ][0];
			}
		}

		return $sizes;
	}

	/**
	 * Returns the width and height of a given thumbnail size.
	 *
	 * @param $size string Thumbnail size name.
	 *
	 * @return array|false Associative array of width and height in pixels. False on invalid size.
	 */
	public function get_thumbnail_dimensions( $size ) {

		switch ( $size ) {
			case 'thumbnail':
			case 'medium':
			case 'large':
				$width  = get_option( $size . '_size_w' );
				$height = get_option( $size . '_size_h' );
				break;

			default:
				if ( empty( $this->additional_image_sizes[ $size ] ) ) {
					return false;
				}

				$width  = $this->additional_image_sizes[ $size ]['width'];
				$height = $this->additional_image_sizes[ $size ]['height'];
		}

		// Just to be safe
		$width  = (int) $width;
		$height = (int) $height;

		return [
			'width'  => $width,
			'height' => $height,
		];
	}

	/**
	 * Fetches the coordinates for a custom crop for a given attachment ID and thumbnail size.
	 *
	 * @param $attachment_id int Attachment ID.
	 * @param $size string Thumbnail size name.
	 *
	 * @return array|false Array of crop coordinates or false if no custom selection set.
	 */
	public function get_coordinates( $attachment_id, $size ) {
		$sizes = (array) get_post_meta( $attachment_id, $this->post_meta, true );

		$coordinates = false;

		if ( empty( $sizes[ $size ] ) ) {
			// Coordinates not explictly set for this size, but is it in a size group? If so, we can use the coordinates
			// from other sizes in the same group, as they are always the same. Happens if a size is added to a group later and hasn't
			// been backfilled in all post meta. Not sure why coords are saved for every size, rather than group, but hey.
			if ( $this->use_ratio_map ) {
				foreach ( $this->image_ratio_map as $ratio => $ratio_sizes ) {
					foreach ( $ratio_sizes as $ratio_size ) {
						if ( $size === $ratio_size ) {
							// Determine if there are any saved coordinates that match the desired $size in the matched ratio group
							$intersect = array_intersect_key( $ratio_sizes, $sizes );

							if ( is_array( $intersect ) && ! empty( $intersect ) ) {
								foreach ( $intersect as $matching_size ) {
									if ( isset( $sizes[ $matching_size ] ) ) {
										$coordinates = $sizes[ $matching_size ];

										break;
									}
								}
							}
						}
					}
				}
			}
		} else {
			$coordinates = $sizes[ $size ];
		}

		return $coordinates;
	}

	/**
	 * Saves the coordinates for a custom crop for a given attachment ID and thumbnail size.
	 *
	 * @param $attachment_id int Attachment ID.
	 * @param $size string Thumbnail size name.
	 * @param $coordinates array Array of coordinates in the format array( x1, y1, x2, y2 )
	 */
	public function save_coordinates( $attachment_id, $size, $coordinates ) {
		$sizes = (array) get_post_meta( $attachment_id, $this->post_meta, true );

		$sizes[ $size ] = $coordinates;

		// save meta for all the related sizes to if were using a ratio map
		if ( $this->use_ratio_map ) {

			$related_sizes = $this->get_related_sizes( $size );

			// add the same meta value to the related sizes
			if ( count( $related_sizes ) ) {
				foreach ( $related_sizes as $related_size ) {
					$sizes[ $related_size ] = $coordinates;
				}
			}
		}

		update_post_meta( $attachment_id, $this->post_meta, $sizes );
	}

	/**
	 * Find the siblings of the passed size so we can apply the coordinates to them too.
	 *
	 * @param $size
	 *
	 * @return array Array of related image size strings
	 */
	public function get_related_sizes( $size ) {

		$related_sizes = [];

		// find out which ratio map the size belongs to
		foreach ( $this->image_ratio_map as $ratio => $ratio_sizes ) {
			foreach ( $ratio_sizes as $ratio_size ) {
				if ( $ratio_size === $size ) {
					$related_sizes = $this->image_ratio_map[ $ratio ];
					break 2;
				}
			}
		}

		return $related_sizes;
	}

	/**
	 * Deletes the coordinates for a custom crop for a given attachment ID and thumbnail size.
	 *
	 * @param $attachment_id int Attachment ID.
	 * @param $size string Thumbnail size name.
	 *
	 * @return bool False on failure (probably no such custom crop), true on success.
	 */
	public function delete_coordinates( $attachment_id, $size ) {
		$sizes = get_post_meta( $attachment_id, $this->post_meta, true );
		if ( ! $sizes ) {
			return false;
		}

		if ( empty( $sizes[ $size ] ) ) {
			return false;
		}

		unset( $sizes[ $size ] );

		// also unset related sizes
		if ( $this->use_ratio_map ) {
			$related_sizes = $this->get_related_sizes( $size );
			if ( count( $related_sizes ) ) {
				foreach ( $related_sizes as $related_size ) {
					unset( $sizes[ $related_size ] );
				}
			}
		}

		return update_post_meta( $attachment_id, $this->post_meta, $sizes );
	}

	/**
	 * Returns the attributes for a given attachment thumbnail. Meant for hooking into image_downsize().
	 *
	 * @param $existing_resize array|false Any existing data. Returned on no action.
	 * @param $attachment_id int Attachment ID.
	 * @param $size string Thumbnail size name.
	 *
	 * @return mixed Array of thumbnail details (URL, width, height, is_intermedite) or the previous data.
	 */
	public function get_thumbnail_url( $existing_resize, $attachment_id, $size ) {
		
		// Named sizes only
		if ( is_array( $size ) ) {
			return $existing_resize;
		}

		$coordinates = $this->get_coordinates( $attachment_id, $size );

		if ( ! $coordinates || ! is_array( $coordinates ) || 4 !== count( $coordinates ) ) {
			return $existing_resize;
		}

		$thumbnail_size = $this->get_thumbnail_dimensions( $size );
		if ( ! $thumbnail_size ) {
			return $existing_resize;
		}

		list( $selection_x1, $selection_y1, $selection_x2, $selection_y2 ) = $coordinates;

		$url = add_query_arg(
			[
				'crop'   =>
					implode(
						'%2C', [
							$selection_x1 . 'px',
							$selection_y1 . 'px',
							( $selection_x2 - $selection_x1 ) . 'px',
							( $selection_y2 - $selection_y1 ) . 'px',
						]
					),
				'resize' => implode(
					'%2C', [
						$thumbnail_size['width'],
						$thumbnail_size['height'],
					]
				),
			],
			wp_get_attachment_url( $attachment_id )
		);

		return [ $url, $thumbnail_size['width'], $thumbnail_size['height'], false ];

	}

}

// initializing the class on init so we can filter the args
add_action(
	'init', function () {
		$GLOBALS['Vip_Go_Thumbnail_Editor'] = new VIP_GO_Thumbnail_Editor;
	}
);
